#!/usr/bin/python2.4
# db_backend.py - Implements DB access for checklib.py and friends
# Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import exceptions, micro
from pysqlite2 import dbapi2 as dbms

def transactional(f):
    def _transactional_wrapper(*args, **kwargs):
        try: 
            res = f(*args, **kwargs)
        except:
            conn.rollback()
            raise
        else:
            conn.commit()
        return res

    return _transactional_wrapper


conn = None
cu = None

@transactional
def _create_schema():
    print "Creating schema"
    cu.execute("""
        CREATE TABLE IF NOT EXISTS packages (
                             pkg_id INTEGER PRIMARY KEY AUTOINCREMENT,
                             pkg_name VARCHAR NOT NULL UNIQUE,
                             pkg_ver VARCHAR NOT NULL,
                             pkg_has_libinfo INT NOT NULL
                            );""")
    cu.execute("""
        CREATE TABLE IF NOT EXISTS libraries (
                              lib_id INTEGER PRIMARY KEY AUTOINCREMENT,
                              pkg_id INTEGER NOT NULL REFERENCES packages(pkg_id),
                              lib_soname VARCHAR,
                              lib_filename VARCHAR NOT NULL,
                              lib_basename VARCHAR NOT NULL
                             );""")
    cu.execute("""
        CREATE TABLE IF NOT EXISTS symbols (
                             sym_id INTEGER PRIMARY KEY AUTOINCREMENT,
                             lib_id INTEGER NOT NULL REFERENCES libraries(lib_id),
                             sym_name VARCHAR NOT NULL,
                             sym_undef INTEGER NOT NULL,
                             sym_ver VARCHAR,
                             sym_verfile VARCHAR
                            );""")
    cu.execute("""
        CREATE TABLE IF NOT EXISTS links (
                              lnk_id INTEGER PRIMARY KEY AUTOINCREMENT,
                              pkg_id INTEGER NOT NULL REFERENCES packages(pkg_id),
                              lib_id INTEGER NOT NULL REFERENCES libraries(lib_id),
                              lnk_filename VARCHAR NOT NULL,
                              lnk_basename VARCHAR NOT NULL
                             );""")

    cu.execute("""
        CREATE TABLE IF NOT EXISTS result (
                             res_id INTEGER PRIMARY KEY AUTOINCREMENT,
                             res_pkg_name VARCHAR NOT NULL,
                             res_pkg_ver VARCHAR NOT NULL,
                             res_result INTEGER NOT NULL,
                             res_flags INTEGER NOT NULL,
                             res_unneeded VARCHAR
                            );""")
    cu.execute("""
        CREATE TABLE IF NOT EXISTS check_queue (
                             chk_id INTEGER PRIMARY KEY AUTOINCREMENT,
                             chk_pkg_name VARCHAR NOT NULL
                            );""")
    cu.execute("""CREATE INDEX IF NOT EXISTS symbols_lib_id_index ON symbols (lib_id);""")
    cu.execute("""CREATE INDEX IF NOT EXISTS symbols_name_index ON symbols (sym_name);""")
    cu.execute("""CREATE INDEX IF NOT EXISTS symbols_name_mix_index ON symbols (sym_name,lib_id);""")
    cu.execute("""CREATE INDEX IF NOT EXISTS libraries_pkg_id_index ON libraries (pkg_id);""")
    cu.execute("""CREATE INDEX IF NOT EXISTS packages_pkg_name_index ON packages (pkg_name);""")

def _check_consistency(local_tbl):
    cu.execute("PRAGMA foreign_key_list('%s')" % local_tbl); 
    fkeys = cu.fetchall()

    for x,y,foreign_tbl, local_id, foreign_id in fkeys:
         cu.execute("select count(*) from %s " % local_tbl+
                    "where %s NOT IN " % local_id +
                    "(select %s from %s);" % (foreign_id, foreign_tbl));
         count = cu.fetchone()[0]
         if count:
             raise int_except.IntDatabaseError("Inconsistent database")

def _check_schema():
    cu.execute("SELECT name FROM SQLITE_MASTER WHERE type = 'table';")
    tables = [i for (i,) in cu.fetchall()]

    if ('packages' not in tables or
        'libraries' not in tables or
        'symbols' not in tables or
        'result' not in tables):
        return False;
    return True;

def init_database(dbpath):
    global conn, cu

    if conn:
        return

    conn = dbms.connect(dbpath)
    cu = conn.cursor()

    if not _check_schema():
        _create_schema()

    _check_consistency('packages')
    _check_consistency('libraries')
    _check_consistency('symbols')
    _check_consistency('result')

@transactional
def update_result(ures):
    if ures.res_unneeded is None:
        unneeded = None
    else:
        unneeded = ','.join(ures.res_unneeded)

    cu.execute("DELETE FROM check_queue WHERE chk_pkg_name = ?", (ures.res_pkg_name,))
    cu.execute("DELETE FROM result WHERE res_pkg_name = ? AND res_pkg_ver = ?;",
               (ures.res_pkg_name, ures.res_pkg_ver))
    cu.execute("INSERT INTO result (res_pkg_name,res_pkg_ver,res_result,res_flags,res_unneeded)" +
               "VALUES (?,?,?,?,?);",
               (ures.res_pkg_name, ures.res_pkg_ver, ures.res_result, ures.res_flags, unneeded))
    ures.id = cu.lastrowid

def get_results():
    cu.execute("SELECT * FROM result;")
    rows = cu.fetchall()
    if len(rows) == 0:
        return None

    results = []
    for row in rows:
        unneeded = row[5]
        row = { 'id': row[0],
                'res_pkg_name': row[1],
                'res_pkg_ver': row[2],
                'res_result': row[3],
                'res_flags': row[4] }
        
        if unneeded is not None:
            unneeded = unneeded.split(',')

        row['res_unneeded'] = unneeded
        results.append(micro.uResult.from_dict(row))

    return results

@transactional
def add_libpkg(pkginfo):
    cu.execute("DELETE FROM packages WHERE pkg_name = ?", (pkginfo.name,))
    cu.execute("INSERT INTO packages (pkg_name,pkg_ver,pkg_has_libinfo) VALUES (?,?,1);",
               (pkginfo.name, pkginfo.ver))

    pkg_id = cu.lastrowid
    lib_mapper = {}
    for lib in pkginfo.get_libraries():
        cu.execute("""INSERT INTO
                      libraries (pkg_id,lib_soname,lib_filename,lib_basename)
                      VALUES (?,?,?,?);""",
                   (pkg_id, lib.soname, lib.pretty_filename, lib.basename))

        lib_id = cu.lastrowid
        lib_mapper[lib.filename] = lib_id
        for sym in lib.get_symbols():
            cu.execute("""INSERT INTO symbols 
                          (lib_id,sym_name,sym_undef,sym_ver,sym_verfile)
                          VALUES (?,?,?,?,?);""",
                       (lib_id, sym.name, sym.undef, sym.ver_name, sym.ver_file))

    for lnk in pkginfo.get_links():
        if lnk.target not in lib_mapper and lnk.target in pkginfo.elf_filenames:
            continue
        try:
            target_lib = lib_mapper[lnk.target]
        except KeyError:
            print lnk.filename, lnk.target
            print lib_mapper
            raise
        cu.execute("""INSERT INTO
                      links (pkg_id, lib_id, lnk_filename, lnk_basename)
                      VALUES (?,?,?,?);""",
                   (pkg_id, target_lib, lnk.pretty_filename, lnk.basename))

    return pkg_id

def get_libpkg(name):
    cu.execute("SELECT pkg_id FROM packages WHERE pkg_name = ? AND pkg_has_libinfo != 0;", (name,))
    rows = cu.fetchall()
    if not rows:
        return None
    return rows[0][0]

@transactional
def update_packages_list(pkgfile):
    cu.execute("DROP TABLE IF EXISTS pkg_import;")
    cu.execute("""
        CREATE TABLE pkg_import (
                             pkg_id INTEGER PRIMARY KEY AUTOINCREMENT,
                             pkg_name VARCHAR NOT NULL UNIQUE,
                             pkg_ver VARCHAR NOT NULL
                            );""")
    for pkgname, pkg_meta in pkgfile.items():
        cu.execute("INSERT INTO pkg_import (pkg_name,pkg_ver) VALUES (?,?)", (pkgname, pkg_meta['Version']))

    cu.execute("""CREATE INDEX IF NOT EXISTS pkg_import_name ON pkg_import (pkg_name);""")
    cu.execute("""CREATE INDEX IF NOT EXISTS pkg_import_ver ON pkg_import (pkg_ver);""")

    cu.execute("""
        DELETE FROM packages WHERE pkg_id IN (
            SELECT l.pkg_id FROM packages AS l, pkg_import AS p 
            WHERE l.pkg_name = p.pkg_name AND l.pkg_ver != p.pkg_ver
        );""")
    cu.execute("""
        DELETE FROM libraries WHERE pkg_id NOT IN (SELECT pkg_id FROM packages);""")
    cu.execute("""
        DELETE FROM symbols WHERE lib_id NOT IN (SELECT lib_id FROM libraries);""")

    cu.execute("""
        DELETE FROM result WHERE res_id IN (
            SELECT res_id FROM result,pkg_import
            WHERE result.res_pkg_name = pkg_import.pkg_name AND 
                  result.res_pkg_ver != pkg_import.pkg_ver);""")

    cu.execute("""
        DELETE FROM result WHERE res_id IN (
            SELECT res_id FROM result,pkg_import
            WHERE result.res_pkg_name = pkg_import.pkg_name AND 
                  result.res_pkg_ver != pkg_import.pkg_ver);""")
    cu.execute("""
        DELETE FROM result WHERE res_pkg_name NOT IN (SELECT pkg_name FROM pkg_import);""")

    cu.execute("""
        INSERT INTO check_queue (chk_pkg_name)
            SELECT pkg_name FROM pkg_import 
                WHERE pkg_name NOT IN (SELECT chk_pkg_name FROM check_queue) AND
                (
                 pkg_name NOT IN (SELECT res_pkg_name FROM result) OR
                 pkg_id IN 
                 (
                  SELECT pkg_id FROM pkg_import,result 
                  WHERE pkg_name = res_pkg_name AND pkg_ver != res_pkg_ver
                 )
                );""")

def get_check_queue():
    cu.execute("SELECT chk_pkg_name FROM check_queue ORDER BY chk_pkg_name;")
    rows = cu.fetchall()
    for row in rows:
        yield row[0]

def all_libs_from_scope(scope):
    qmarks = ','.join(['?'] * len(scope))

    cu.execute("""SELECT pkg_id, lib_filename, lib_basename, lib_id 
                  FROM libraries WHERE pkg_id IN (%s);""" % qmarks,
                  scope)
    libs = cu.fetchall()

    cu.execute("""SELECT pkg_id, lnk_filename, lnk_basename, lib_id
                  FROM links WHERE pkg_id IN (%s);""" % qmarks,
                  scope)
    libs += cu.fetchall()

    names = ['pkg_id', 'filename', 'basename', 'lib_id']
    libs = [dict(zip(names, lib)) for lib in libs]

    return libs

def pkg_names_from_pkg_ids(scope):
    qmarks = ','.join(['?'] * len(scope))
    cu.execute("""SELECT pkg_id, pkg_name FROM packages WHERE pkg_id IN (%s);""" % qmarks, scope)
    return cu.fetchall()


def pkg_ids_from_lib_ids(lib_ids):
    qmarks = ','.join(['?'] * len(lib_ids))
    cu.execute("""SELECT pkg_id FROM libraries WHERE lib_id IN (%s);""" % qmarks, lib_ids)

    return [i for (i,) in cu.fetchall()]


@transactional
def insert_binary(bin, load_libs):
    cu.execute("""DROP TABLE IF EXISTS binary;""")
    cu.execute("""DROP TABLE IF EXISTS bin_symbol;""")
    cu.execute("""DROP TABLE IF EXISTS bin_libs;""")
    cu.execute("""
        CREATE TABLE binary (
                       bin_id INTEGER PRIMARY KEY AUTOINCREMENT,
                       bin_filename VARCHAR NOT NULL
                      );""")
    cu.execute("""
        CREATE TABLE bin_symbol (
                       bsym_id INTEGER PRIMARY KEY AUTOINCREMENT,
                       bin_id INTEGER NOT NULL REFERENCES binary(bin_id),
                       sym_name VARCHAR NOT NULL
                      );""")
    cu.execute("""
        CREATE TABLE bin_libs (
                       blib_id INTEGER PRIMARY KEY AUTOINCREMENT,
                       bin_id INTEGER NOT NULL REFERENCES binary(bin_id),
                       lib_id INTEGER NOT NULL REFERENCES libraries(lib_id)
                      );""")

    cu.execute("""INSERT INTO binary (bin_filename) VALUES (?);""", (bin.filename,))
    bin_id = cu.lastrowid

    for lib_id in load_libs:
        cu.execute("""INSERT INTO bin_libs (bin_id, lib_id) VALUES (?,?);""", (bin_id, lib_id))

    sym_map = {}
    for symbol in bin.get_symbols():
        cu.execute("""INSERT INTO bin_symbol (bin_id, sym_name) VALUES (?,?);""", (bin_id, symbol.name))
        symbol.id = cu.lastrowid
        sym_map[symbol.id] = symbol

    return bin_id, sym_map

def find_needed_libs(bin_id):
    cu.execute("""
                SELECT DISTINCT lib_id FROM bin_symbol NATURAL JOIN symbols WHERE
                    lib_id IN (SELECT lib_id FROM bin_libs)
                    AND sym_undef = 0
            ;""", [])
    return [i for (i,) in cu.fetchall()]

def find_unresolved_syms(bin_id):
    cu.execute("""
            SELECT bsym_id FROM bin_symbol WHERE sym_name NOT IN (
                SELECT sym_name FROM symbols NATURAL JOIN bin_libs
            );""")
    return [i for (i,) in cu.fetchall()]
