#!/usr/bin/python
# elfobject.py - Get information out of ELF binaries/libraries
#                Used by checklib.py and friends.
# Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import os
import elf
import symbol
import objinfo

class ElfError(RuntimeError):
    pass

class NotALibrary(ElfError):
    pass

class Library:
    def __init__(self, filename, strip_dir_len=0):
        self.filename = filename
        self.pretty_filename = filename[strip_dir_len:]
        self.basename = os.path.basename(filename)

        if elf.elf_type(filename) != elf.ET_DYN:
            raise NotALibrary("Not a library")

    def get_soname(self):
        self.soname = None

        for line in objinfo.get_dynamic(self.filename):
            data = line.strip().split(':', 2)
            assert data[0] == 'DYN'

            if data[1] == 'SONAME':
                assert not self.soname
                self.soname = data[2]

    def get_symbols(self):
        for line in objinfo.get_exported(self.filename):
            sym = symbol.ExportSym(self.filename, line.strip())
            if not sym.undef:
                yield sym

class Binary:
    def __init__(self, filename, strip_dir_len=0):
        self.filename = filename
        self.pretty_filename = filename[strip_dir_len:]
        self.basename = os.path.basename(filename)

    def get_dynamic(self):
        self.needed = []
        self.rpath = []
        self.runpath = []
        self.textrel = False

        for line in objinfo.get_dynamic(self.filename):
            data = line.strip().split(':', 2)
            assert data[0] == 'DYN'

            if data[1] == 'NEEDED':
                self.needed.append(data[2])
            elif data[1] == 'RPATH':
                assert not self.rpath
                self.rpath = data[2].split(':')
            elif data[1] == 'RUNPATH':
                assert not self.runpath
                self.runpath = data[2].split(':')
            elif data[1] == 'TEXTREL':
                self.textrel = True

        # if runpath is present, rpath is ignored
        if self.runpath:
            self.rpath = []

    def get_symbols(self):
        symindex_cache = {}
        for line in objinfo.get_relocations(self.filename):
            sym = symbol.RelocSym(self.filename, line.strip())
            if sym.index not in symindex_cache:
                # We haven't seen this particular symbol yet
                symindex_cache[sym.index] = None
                yield sym

def main():
    for filename in sys.argv[1:]:
        if not elf.elf_type(filename):
            continue

        bin = Binary(filename)
        bin.get_dynamic()
        for sym in bin.get_symbols():
            print sym

        lib = Library(filename)
        for sym in lib.get_symbols():
            print sym

if __name__ == '__main__':
    main()
