/* objinfo.c - print miscellaneous information about ELF files
   This file is loosely based on readelf.c from the Red Hat elfutils
   Copyright (C) 2006 Christian Aichinger <Greek0@gmx.net>
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006 Red Hat, Inc.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
   */

#include <argp.h>
#include <fcntl.h>
#include <gelf.h>
#include <inttypes.h>
#include <libelf.h>
#include <elfutils/libebl.h>
#include <libintl.h>
#include <locale.h>
#include <mcheck.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>

#include "utils.h"
#include "commands.h"
#include "output.h"


/* Definitions of arguments for argp functions.  */
static const struct argp_option options[] =
{
  { NULL, 0, NULL, 0, "Commands", 0 },
  { NULL, 'A', NULL, 0, "Show overview", 0 },
  { NULL, 'B', NULL, 0, "Show relocations", 0 },
  { NULL, 'C', NULL, 0, "Show hashtable", 0 },
  { NULL, 'D', NULL, 0, "Show version info", 0 },
  { NULL, 'E', NULL, 0, "Show dynamic section", 0 },
  { NULL, 0, NULL, 0, "Output formats", 0 },
  { NULL, 'n', NULL, 0, "Don't print headers", 0 },
  { NULL, 'm', NULL, 0, "Machine-readable output", 0 },
  { NULL, 0, NULL, 0, NULL, 0 }
};

/* Short description of program.  */
static const char doc[] = "\
displays information about ELF files";

/* Strings for arguments in help texts.  */
static const char args_doc[] = "[FILE...]";

/* Prototype for option handler.  */
static error_t parse_opt (int key, char *arg, struct argp_state *state);

/* Data structure to communicate with argp functions.  */
static struct argp argp =
{
  options, parse_opt, args_doc, doc, NULL, NULL, NULL
};

/* Print symbols in file named FNAME.  */
static bool process_file (const char *fname);

/* The info to display about the file */
const uint32_t do_overview    = 1<<0;
const uint32_t do_relocations = 1<<1;
const uint32_t do_hashtable   = 1<<2;
const uint32_t do_versioninfo = 1<<3;
const uint32_t do_dynamic     = 1<<4;
uint32_t command = 0;

int
main (int argc, char *argv[])
{
  int remaining;
  bool result = true;

  /* Make memory leak detection possible.  */
  mtrace ();

  /* We use no threads here which can interfere with handling a stream.  */
  __fsetlocking (stdin, FSETLOCKING_BYCALLER);
  __fsetlocking (stdout, FSETLOCKING_BYCALLER);
  __fsetlocking (stderr, FSETLOCKING_BYCALLER);

  /* Parse and process arguments.  */
  argp_parse (&argp, argc, argv, 0, &remaining, NULL);

  /* Tell the library which version we are expecting.  */
  elf_version (EV_CURRENT);

  if (remaining == argc)
    /* The user didn't specify a name so we use a.out.  */
    result = process_file ("a.out");
  else
    /* Process all the remaining files.  */
    do
      result &= process_file (argv[remaining]);
    while (++remaining < argc);

  if (result)
      return EXIT_SUCCESS;
  return EXIT_FAILURE;
}


/* Handle program arguments.  */
static error_t
parse_opt (int key, char *arg,
	   struct argp_state *state __attribute__ ((unused)))
{
  switch (key)
    {
    case 'A':
      command |= do_overview;
      break;

    case 'B':
      command |= do_relocations;
      break;

    case 'C':
      command |= do_hashtable;
      break;

    case 'D':
      command |= do_versioninfo;
      break;

    case 'E':
      command |= do_dynamic;
      break;

    case 'n':
      output_format |= fmt_noheaders;
      break;

    case 'm':
      output_format |= fmt_machine;
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

/* Open the file and determine the type.  */
static bool process_file (const char *fname) {
    bool result = false;
    int fd = open (fname, O_RDONLY);
    if (fd == -1) {
        error("cannot open '%s'\n", fname);
        return 1;
    }
  
    /* Now get the ELF descriptor.  */
    Elf *elf = elf_begin (fd, ELF_C_READ_MMAP, NULL);
    if (!elf) {
        error ("File format not recognized: '%s'\n", fname);
        goto out_fd;
    }

    if (elf_kind(elf) != ELF_K_ELF) {
        error ("File format not recognized: '%s'\n", fname);
        goto out_elf;
    }

    Ebl *ebl = ebl_openbackend(elf);
    if (ebl == NULL) {
        error ("cannot create EBL handle for '%s'\n", fname);
        goto out_elf;
    }

    UElf_File* efile = file_new(elf, ebl, fname);
    if (!efile) {
        error ("cannot create UElf_File object for '%s'\n", fname);
        goto out_ebl;
    }

    result = true;
    printf("%s\n", fname);
    if (command & do_overview)
        show_overview(efile);
    if (command & do_relocations)
        show_relocations(efile);
    if (command & do_hashtable)
        show_hashtable(efile);
    if (command & do_versioninfo)
        show_versioninfo(efile);
    if (command & do_dynamic)
        show_dynamic(efile);

    file_free(efile);
out_ebl:
    ebl_closebackend(ebl);

out_elf:
    if (elf_end(elf) != 0)
        internal_error("Couldn't destroy elf object for '%s'\n", fname);
out_fd:
    if (close(fd) != 0)
      internal_error ("Cannot close '%s'\n", fname);

    return result;
}
