/* output.c - functions for outputting internal representations of ELF info
   This file is loosely based on readelf.c from the Red Hat elfutils
   Copyright (C) 2006 Christian Aichinger <Greek0@gmx.net>
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006 Red Hat, Inc.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
   */

#include <gelf.h>
#include <inttypes.h>
#include <libelf.h>
#include <elfutils/libebl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "utils.h"
#include "output.h"

uint32_t output_format = 0;

#define do_vprintf(stream, format)     \
    do {                               \
        va_list arg;                   \
        va_start (arg, format);        \
        vfprintf(stream, format, arg); \
        va_end (arg);                  \
    } while (0)

void _error(const char *file, unsigned int line, const char *format, ...) {
    fflush(stdout);
    if (file)
        fprintf(stderr, "E: (at %s:%d) ", file, line);
    else
        fprintf(stderr, "E: ");
    do_vprintf(stderr, format);
}

void _debug(const char *file, unsigned int line, const char *format, ...) {
    fflush(stdout);
    fprintf(stderr, "D: (at %s:%d) ", file, line);
    do_vprintf(stderr, format);
}

void print_header(const char *format, ...) {
    if ((output_format & fmt_noheaders) == 0)
        do_vprintf(stdout, format);
}

static const char* get_symbol_vis_name(unsigned char vis, char* buf, size_t len) {
    switch (vis) {
        case 0: return "DEFAULT";
        case 1: return "INTERNAL";
        case 2: return "HIDDEN";
        case 3: return "PROTECTED";
        default:
                snprintf(buf, len, "%d", vis);
                return buf;
    }
}

#define SEP ":"
#define fmt_printf(fmt1, fmt2, ...)         \
    do {                                    \
        if (output_format & fmt_machine)    \
            printf(fmt1, ## __VA_ARGS__);   \
        else                                \
            printf(fmt2, ## __VA_ARGS__);   \
    } while (0)


void print_reloc(UElf_File *efile, GElf_Rel *rel) {
    GElf_Word type = GELF_R_TYPE(rel->r_info);
    bool type_match = ebl_reloc_type_check (efile->ebl, type);
    
    char buf[128];
    const char *reloc_type = type_match
        ? ebl_reloc_type_name(efile->ebl, type, buf, sizeof (buf))
        : "<INVALID_RELOC>";

    fmt_printf(
            "REL"SEP"%#010llx"SEP"%s"SEP,
            "  %#010" PRIx64 " %-25s ",
            rel->r_offset, reloc_type);
}

void print_sym(UElf_File *efile, GElf_Word index, UElf_Symbol *info) {
    if (!(info->flags & SIF_SYMBOL)) {
        printf("<INVALID_SYMBOL_%d>", index);
        return;
    }

    if (info->type == STT_SECTION && !(info->flags & SIF_SEC_VALID)) {
        printf("<INVALID_SECTION_%d>", index);
        return;
    }

    char buf1[120], buf2[120], buf3[120], buf4[120];
    const char *bindingname = ebl_symbol_binding_name(efile->ebl, info->bind, buf1, sizeof(buf1));
    const char *typename = ebl_symbol_type_name(efile->ebl, info->type, buf2, sizeof(buf2));
    const char *visibility = get_symbol_vis_name(
            GELF_ST_VISIBILITY(info->sym.st_other),
            buf3, sizeof(buf3));
    const char *section = ebl_section_name(efile->ebl, info->sym.st_shndx, 0,
            buf4, sizeof(buf4), efile->scnnames, efile->shnum);

    char *psec;
    if (info->sym.st_shndx > 0 && info->sym.st_shndx < SHN_LORESERVE) {
        psec = (char *)section;
    } else {
        psec = alloca(strlen(section) + 10);
        sprintf(psec, "SHN_%s", section);
    }

    fmt_printf(
            "SYM"SEP"STT_%s"SEP"STB_%s"SEP"STV_%s"SEP"%s"SEP"%#010llx"SEP"%#010llx"SEP"%04d"SEP"%s"SEP,
            "  STT_%-8s STB_%-8s STV_%-10s %-12s 0x%08llx %05lld -- [%04d] %-35s",
            typename, bindingname, visibility, psec, info->sym.st_value,
            info->sym.st_size, info->index, info->name);
}

void print_verinfo(UElf_File *efile, GElf_Word index, UElf_VersionInfo *verinfo) {
    UElf_SymVersion sv;
    if (verinfo) {
        versym_fill(efile, verinfo, index, &sv);
        fmt_printf(
            "VS"SEP"%04u"SEP"%d"SEP"%d"SEP"%s"SEP"%s"SEP"%s\n",
            "  veridx: %04u, type: %d, flags: %d, %s, %s, %s\n",
            sv.index, sv.meta->type,
            sv.meta->type == VMT_VERDEF ? sv.meta->def.vd.vd_flags : 
            sv.meta->type == VMT_VERNEED ? sv.meta->need.vna.vna_flags : -1,
            sv.hidden ? "is  hidden"      : "not hidden", 
            sv.meta->vername, 
            sv.meta->filename);
    } else
        fmt_printf( "NOVS\n", "  No version info available\n");
}

void print_section(UElf_File *efile, unsigned int index, GElf_Shdr *shdr) {
    char buf[120];
    const char* name = elf_strptr(efile->elf, efile->shstrndx, shdr->sh_name);
    const char* type = ebl_section_type_name(efile->ebl, shdr->sh_type, buf, sizeof (buf));
    Elf32_Word addr = shdr->sh_addr;
    Elf32_Word size = shdr->sh_size;
    char flags[sizeof(shdr->sh_flags)*8+1];

#define MAX_SHF SHF_TLS
#define set_flag(flags, shdrflags, name, chr) \
    do { flags[ffsl(name)-1] = (shdrflags & name) ? chr : '-'; } while (0)

    memset(flags, '-', sizeof(flags));

    set_flag(flags, shdr->sh_flags, SHF_WRITE, 'w');
    set_flag(flags, shdr->sh_flags, SHF_ALLOC, 'a');
    set_flag(flags, shdr->sh_flags, SHF_EXECINSTR, 'x');
    set_flag(flags, shdr->sh_flags, SHF_MERGE, 'm');
    set_flag(flags, shdr->sh_flags, SHF_STRINGS, 's');
    set_flag(flags, shdr->sh_flags, SHF_INFO_LINK, 'i');
    set_flag(flags, shdr->sh_flags, SHF_LINK_ORDER, 'l');
    set_flag(flags, shdr->sh_flags, SHF_OS_NONCONFORMING, 'N');
    set_flag(flags, shdr->sh_flags, SHF_GROUP, 'G');
    set_flag(flags, shdr->sh_flags, SHF_TLS, 'T');

    flags[ffsl(MAX_SHF)] = '\0';
  
    fmt_printf(
            "SEC"SEP"%02u"SEP"%s"SEP"%#010x"SEP"%05x"SEP"%s"SEP"%s\n",
            "[%02u] %-30s at %#010x, size %05x, flags %s, type %s\n",
            index, name, addr, size, flags, type);
}

struct flags {
  int mask;
  const char *str;
};

static const struct flags dt_flags[] = {
    { DF_ORIGIN, "ORIGIN" },
    { DF_SYMBOLIC, "SYMBOLIC" },
    { DF_TEXTREL, "TEXTREL" },
    { DF_BIND_NOW, "BIND_NOW" },
    { DF_STATIC_TLS, "STATIC_TLS" },
    { 0 }
};
static const struct flags dt_flags_1[] = {
    { DF_1_NOW, "NOW" },
    { DF_1_GLOBAL, "GLOBAL" },
    { DF_1_GROUP, "GROUP" },
    { DF_1_NODELETE, "NODELETE" },
    { DF_1_LOADFLTR, "LOADFLTR" },
    { DF_1_INITFIRST, "INITFIRST" },
    { DF_1_NOOPEN, "NOOPEN" },
    { DF_1_ORIGIN, "ORIGIN" },
    { DF_1_DIRECT, "DIRECT" },
    { DF_1_TRANS, "TRANS" },
    { DF_1_INTERPOSE, "INTERPOSE" },
    { DF_1_NODEFLIB, "NODEFLIB" },
    { DF_1_NODUMP, "NODUMP" },
    { DF_1_CONFALT, "CONFALT" },
    { DF_1_ENDFILTEE, "ENDFILTEE" },
    { DF_1_DISPRELDNE, "DISPRELDNE" },
    { DF_1_DISPRELPND, "DISPRELPND" },
    { 0 }
};
static const struct flags dt_feature_1[] = {
    { DTF_1_PARINIT, "PARINIT" },
    { DTF_1_CONFEXP, "CONFEXP" },
    { 0 }
};
static const struct flags dt_posflag_1[] = {
    { DF_P1_LAZYLOAD, "LAZYLOAD" },
    { DF_P1_GROUPPERM, "GROUPPERM" },
    { 0 }
};

static void fill_flags(GElf_Xword d_val, const struct flags *flags_start, char* buf, int len0) {
  const struct flags *ptr = flags_start;
  const char* seperator = "";

  char *str = buf;
  int len = len0;

  while (ptr && ptr->mask && ptr->str && len > 0) {
      if (d_val & ptr->mask) {
          int res = snprintf(str, len, "%s%s", seperator, ptr->str);
          str += res;
          len -= res;

          seperator = "|";
          d_val &= ~ptr->mask;
      }
      ptr++;
  }
  if (d_val != 0 && len > 0)
      snprintf(str, len, "%s%#010" PRIx64, seperator, d_val);
}

static void print_flags(GElf_Xword d_val, const struct flags *flags_start) {
    char buf[512];
    fill_flags(d_val, flags_start, buf, sizeof(buf));
    printf("%s\n", buf);
}

void print_dynamic_entry(UElf_File *efile, UElf_Section *section, GElf_Dyn *dyn) {
    char buf[120];
    const char *tag_name = ebl_dynamic_tag_name(efile->ebl, dyn->d_tag, buf, sizeof(buf));

    fmt_printf("DYN"SEP"%s"SEP, "  %-17s ", tag_name);
    switch (dyn->d_tag) {
        case DT_NULL:
        case DT_DEBUG:
        case DT_BIND_NOW:
        case DT_TEXTREL:
            /* No further output.  */
            fputc_unlocked ('\n', stdout);
            break;
        case DT_NEEDED:
            printf("%s\n",
                    elf_strptr(efile->elf, section->shdr->sh_link, dyn->d_un.d_val));
            break;
        case DT_SONAME:
            printf("%s\n",
          	    elf_strptr(efile->elf, section->shdr->sh_link, dyn->d_un.d_val));
            break;
          
        case DT_RPATH:
            printf("%s\n",
                    elf_strptr(efile->elf, section->shdr->sh_link, dyn->d_un.d_val));
            break;
        case DT_RUNPATH:
            printf("%s\n",
          	    elf_strptr(efile->elf, section->shdr->sh_link, dyn->d_un.d_val));
            break;
        case DT_PLTRELSZ:
        case DT_RELASZ:
        case DT_STRSZ:
        case DT_RELSZ:
        case DT_RELAENT:
        case DT_SYMENT:
        case DT_RELENT:
        case DT_PLTPADSZ:
        case DT_MOVEENT:
        case DT_MOVESZ:
        case DT_INIT_ARRAYSZ:
        case DT_FINI_ARRAYSZ:
        case DT_SYMINSZ:
        case DT_SYMINENT:
        case DT_GNU_CONFLICTSZ:
        case DT_GNU_LIBLISTSZ:
            fmt_printf(
                    "%" PRId64 "\n",
                    "%" PRId64 " (bytes)\n",
                    dyn->d_un.d_val);
            break;
        case DT_VERDEFNUM:
        case DT_VERNEEDNUM:
        case DT_RELACOUNT:
        case DT_RELCOUNT:
            printf("%" PRId64 "\n", dyn->d_un.d_val);
            break;
        case DT_PLTREL:
            puts(ebl_dynamic_tag_name(efile->ebl, dyn->d_un.d_val, NULL, 0));
            break;
        case DT_FLAGS:
            print_flags(dyn->d_un.d_val, dt_flags);
            break;
        case DT_FLAGS_1:
            print_flags(dyn->d_un.d_val, dt_flags_1);
            break;
        case DT_FEATURE_1:
            print_flags(dyn->d_un.d_val, dt_feature_1);
            break;
        case DT_POSFLAG_1:
            print_flags(dyn->d_un.d_val, dt_posflag_1);
            break;
        default:
            printf("%#010" PRIx64 "\n", dyn->d_un.d_val);
            break;
    }
}

void print_verneed(UElf_File *efile, UElf_Section *section, unsigned int offset, GElf_Verneed *need) {
    fmt_printf("VN"SEP"P"SEP"%#06x"SEP"%hu"SEP"%s"SEP"%hu\n",
               "  %#06x: Version: %hu File: %s  Cnt: %hu\n",
	        offset,
                (unsigned short int)need->vn_version,
	        elf_strptr(efile->elf, section->shdr->sh_link, need->vn_file),
	        (unsigned short int) need->vn_cnt);
}

void print_verneedaux(UElf_File *efile, UElf_Section *section, unsigned int offset, GElf_Vernaux *aux) {
    fmt_printf("VN"SEP"C"SEP"%#06x"SEP"%s"SEP"%x"SEP"%hu\n",
               "      %#06x: Name: %-20s  Flags: %x  Version: %hu\n",
	  	offset,
	  	elf_strptr(efile->elf, section->shdr->sh_link, aux->vna_name),
	  	aux->vna_flags,
	  	(unsigned short int)aux->vna_other);
}

void print_verdef(UElf_File *efile, UElf_Section *section,
        unsigned int offset, GElf_Verdef *def, GElf_Verdaux *aux) {

    fmt_printf("VD"SEP"P"SEP"%#06x"SEP"%hd"SEP"%x"SEP"%hd"SEP"%hd"SEP"%s\n",
               "  %#06x: Version: %hd  Flags: %x  Index: %hd  Cnt: %hd,  Name: %s\n",
	        offset,
                (unsigned short int)def->vd_version,
                def->vd_flags,
                def->vd_ndx,
	        (unsigned short int) def->vd_cnt,
	        elf_strptr(efile->elf, section->shdr->sh_link, aux->vda_name));
}

void print_verdefaux(UElf_File *efile, UElf_Section *section, unsigned int offset, GElf_Verdaux *aux) {
    fmt_printf("VD"SEP"C"SEP"%#06x"SEP"%s\n",
               "      %#06x: Parent: %s\n",
	        offset,
		elf_strptr(efile->elf, section->shdr->sh_link, aux->vda_name));
}
