/* relocations.c - info about relocation entries (needed syms) in ELF files
   This file is loosely based on readelf.c from the Red Hat elfutils
   Copyright (C) 2006 Christian Aichinger <Greek0@gmx.net>
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006 Red Hat, Inc.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
   */

#include <gelf.h>
#include <inttypes.h>
#include <libelf.h>
#include <elfutils/libebl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <string.h>

#include "utils.h"
#include "output.h"

static void handle_relocs_rel(UElf_File *efile, UElf_Section *section) {
    int nentries = section->shdr->sh_entsize ? (section->shdr->sh_size / section->shdr->sh_entsize) : 0;
  
    section_fill_data(efile, section);
    if (!section->data) {
        warning("Can't get data for REL section\n"); 
        return;
    }
  
    UElf_Section sec_sym = UElf_Section_INIT;
    section_from_index(efile, section->shdr->sh_link, &sec_sym);
    section_fill_data(efile, &sec_sym);
  
    UElf_Section sec_dest = UElf_Section_INIT;
    section_from_index(efile, section->shdr->sh_info, &sec_dest);

    if (sec_sym.shdr == NULL || sec_sym.data == NULL || sec_dest.shdr == NULL) {
        warning("Invalid symbol table at offset %#0" PRIx64 "\n", section->shdr->sh_offset);
        return;
    }
  
    char *dataname = section->shdr->sh_info ?
                     elf_strptr(efile->elf, efile->shstrndx, sec_dest.shdr->sh_name) : "";
  
    print_header("Relocation section [%2u] '%s' for section [%2u] '%s' at offset %#0" PRIx64 ", # of entries: %d\n",
            (unsigned int) elf_ndxscn(section->scn),
            elf_strptr(efile->elf, efile->shstrndx, section->shdr->sh_name),
            (unsigned int) section->shdr->sh_info, 
            dataname,
            section->shdr->sh_offset,
            nentries);
  
    UElf_VersionInfo *verinfo = verinfo_new(efile);

    for (int cnt = 0; cnt < nentries; ++cnt) {
        GElf_Rel relmem;
        GElf_Rel *rel = gelf_getrel(section->data, cnt, &relmem);
        if (rel == NULL)
            continue;
  
        UElf_Symbol info;
        resolve_symbol(efile, &sec_sym, GELF_R_SYM(rel->r_info), &info);
        GElf_Word symindex = GELF_R_SYM (rel->r_info);
      
        print_reloc(efile, rel);
        print_sym(efile, symindex, &info);
        print_verinfo(efile, symindex, verinfo);
    }

    verinfo_free(verinfo);
}

void
show_relocations (UElf_File *efile) {
    UElf_Section section = UElf_Section_INIT;
    while (section_find(efile, SHT_REL, &section))
        handle_relocs_rel(efile, &section);

    section = UElf_Section_INIT;
    while (section_find(efile, SHT_RELA, &section))
        warning("SHT_RELA section found, skipping\n");
}
