/* versioning.c - info about symbol versioning entries in ELF files
   This file is loosely based on readelf.c from the Red Hat elfutils
   Copyright (C) 2006 Christian Aichinger <Greek0@gmx.net>
   Copyright (C) 2000, 2001, 2002, 2003, 2004, 2005, 2006 Red Hat, Inc.

   This program is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License as published by the Free Software
   Foundation; version 2 of the License.

   This program is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along with
   this program; if not, write to the Free Software Foundation, Inc., 51
   Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
   */

#include <gelf.h>
#include <inttypes.h>
#include <libelf.h>
#include <elfutils/libebl.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "utils.h"
#include "versioniter.h"
#include "output.h"

static void fill_version_names(UElf_File *efile, UElf_VersionInfo *verinfo) {
    if (verinfo->sec_def.scn) {
        UElf_Verdef_Iter iter = UElf_Verdef_Iter_INIT;
        while (iter_verdef(efile, &verinfo->sec_def, &iter)) {
            while (iter_verdefaux(efile, &verinfo->sec_def, &iter)) {
                GElf_Half index = iter.def->vd_ndx & 0x7fff;

                if (iter.auxcount == 1) {
                    UElf_VersionMeta *meta = malloc(sizeof(UElf_VersionMeta));
                    sarray_insert(verinfo->meta, index, meta);

                    meta->vername = elf_strptr(efile->elf, verinfo->sec_def.shdr->sh_link, iter.aux->vda_name);
                    meta->filename = NULL;

                    meta->type = VMT_VERDEF;
                    memcpy(&meta->def.vd, iter.def, sizeof(meta->def.vd));
                    memcpy(&meta->def.vda, iter.aux, sizeof(meta->def.vda));
                }
                // No idea what the second verdaux entries mean here. The spec isn't clear either :-/
                // Uncomment to see them.
                //debug(stderr, "verdef entry #%d, idx %d, aux #%d -> %s\n",
                //      iter.count, index, iter.auxcount, 
                //      elf_strptr(efile->elf, verinfo->sec_def.shdr->sh_link, iter.aux->vda_name));
            }
        }
    }

    if (verinfo->sec_need.scn) {
        UElf_Verneed_Iter iter = UElf_Verneed_Iter_INIT;
        while (iter_verneed(efile, &verinfo->sec_need, &iter)) {
            while (iter_verneedaux(efile, &verinfo->sec_need, &iter)) {
                GElf_Half index = iter.aux->vna_other & 0x7fff;
                
                UElf_VersionMeta *meta = malloc(sizeof(UElf_VersionMeta));
                sarray_insert(verinfo->meta, index, meta);

                meta->vername = elf_strptr(efile->elf, verinfo->sec_need.shdr->sh_link, iter.aux->vna_name);
                meta->filename = elf_strptr(efile->elf, verinfo->sec_need.shdr->sh_link, iter.need->vn_file);

                meta->type = VMT_VERNEED;
                memcpy(&meta->need.vn, iter.need, sizeof(meta->need.vn));
                memcpy(&meta->need.vna, iter.aux, sizeof(meta->need.vna));
	    }
        }
    }
}

UElf_VersionInfo* verinfo_new(UElf_File *efile) {
    UElf_VersionInfo *verinfo = malloc(sizeof(UElf_VersionInfo));
    verinfo->meta = sarray_new();

    verinfo->sec_sym = UElf_Section_INIT;
    section_find(efile, SHT_GNU_versym, &verinfo->sec_sym);
    section_fill_data(efile, &verinfo->sec_sym);
    if (!verinfo->sec_sym.scn || !verinfo->sec_sym.data) {
        sarray_free(verinfo->meta);
        free(verinfo);
        return NULL;
    }

    verinfo->sec_def = UElf_Section_INIT;
    section_find(efile, SHT_GNU_verdef, &verinfo->sec_def);
    section_fill_data(efile, &verinfo->sec_def);

    verinfo->sec_need = UElf_Section_INIT;
    section_find(efile, SHT_GNU_verneed, &verinfo->sec_need);
    section_fill_data(efile, &verinfo->sec_need);

    verinfo->sec_strings = UElf_Section_INIT;
    section_from_index(efile, verinfo->sec_sym.shdr->sh_link, &verinfo->sec_strings);

    fill_version_names(efile, verinfo);

    return verinfo;
}

void verinfo_free(UElf_VersionInfo *verinfo) {
    if (!verinfo)
        return;

    sarray_entry *entry = sarray_first(verinfo->meta);
    while (entry) {
        free(sarray_value(entry));
        entry = sarray_next(entry);
    }

    sarray_free(verinfo->meta);
    free(verinfo);
}

static const UElf_VersionMeta meta0 = {
    .type = VMT_UNKNOWN,
    .vername = "<local>",
    .filename = NULL
};
static const UElf_VersionMeta meta1 = {
    .type = VMT_UNKNOWN,
    .vername = "<global>",
    .filename = NULL
};
static const UElf_VersionMeta metaff01 = {
    .type = VMT_UNKNOWN,
    .vername = "<eleminate>",
    .filename = NULL
};
static const UElf_VersionMeta metaNULL = {
    .type = VMT_UNKNOWN,
    .vername = NULL,
    .filename = NULL
};

void versym_fill(UElf_File *efile, UElf_VersionInfo* verinfo, GElf_Word index, UElf_SymVersion *sv) {
    sv->sym = gelf_getversym(verinfo->sec_sym.data, index, &sv->symmem);
    if (!sv->sym)
        internal_error("Couldn't get version info for '%s'\n", efile->fname);

    sv->hidden = *sv->sym & 0x8000;
    sv->index = *sv->sym & 0x7fff;
    switch (sv->index) {
        case 0:
            sv->meta = &meta0;
            break;
        case 1:
            sv->meta = &meta1;
            break;
        case 0xff01:
            sv->meta = &metaff01;
            break;
        default:
            sv->meta = sarray_find(verinfo->meta, sv->index);
            if (!sv->meta)
                sv->meta = &metaNULL;
    }
}

static void handle_versym(UElf_File *efile, UElf_Section *section) {
    UElf_Section sec_link = UElf_Section_INIT;
    section_from_index(efile, section->shdr->sh_link, &sec_link);
    if (sec_link.shdr == NULL) {
        warning("Couldn't get link for GNU_versym section at %#0" PRIx64 "\n", section->shdr->sh_offset);
        return;
    }

    print_header("Versym: [%2u] '%s', nentries: %d, Addr: %#010" PRIx64 ", Off: %#08" PRIx64 ", Lnk: [%2u] '%s'\n",
	  (unsigned int) elf_ndxscn(section->scn),
	  elf_strptr (efile->elf, efile->shstrndx, section->shdr->sh_name),
	  (int) (section->shdr->sh_size / section->shdr->sh_entsize),
	  section->shdr->sh_addr,
	  section->shdr->sh_offset,
	  (unsigned int) section->shdr->sh_link,
	  elf_strptr (efile->elf, efile->shstrndx, sec_link.shdr->sh_name));
}

static void handle_verneed(UElf_File *efile, UElf_Section *section) {
    section_fill_data(efile, section);
    if (!section->data) {
        warning("Can't get data for GNU_verneed section\n"); 
        return;
    }

    UElf_Section sec_str = UElf_Section_INIT;
    section_from_index(efile, section->shdr->sh_link, &sec_str);
    if (sec_str.shdr == NULL) {
        warning("Couldn't get strings for GNU_verneed section at %#0" PRIx64 "\n", section->shdr->sh_offset);
        return;
    }

    print_header("Verneeds: [%2u] '%s', n_ent: %d, Addr: %#010" PRIx64 ", Offset: %#08llx, Lnk: [%2u] '%s'\n",
	    (unsigned int) elf_ndxscn(section->scn),
	    elf_strptr(efile->elf, efile->shstrndx, section->shdr->sh_name),
            section->shdr->sh_info,
	    section->shdr->sh_addr,
	    section->shdr->sh_offset,
	    (unsigned int) section->shdr->sh_link,
	    elf_strptr(efile->elf, efile->shstrndx, sec_str.shdr->sh_name));


    UElf_Verneed_Iter iter = UElf_Verneed_Iter_INIT;
    while (iter_verneed(efile, section, &iter)) {
        print_verneed(efile, section, iter.offset, iter.need);
        
        while (iter_verneedaux(efile, section, &iter))
            print_verneedaux(efile, section, iter.auxoffset, iter.aux);
    }
}

static void handle_verdef(UElf_File *efile, UElf_Section *section) {
    section_fill_data(efile, section);
    if (!section->data) {
        warning("Can't get data for GNU_verneed section\n"); 
        return;
    }

    UElf_Section sec_str = UElf_Section_INIT;
    section_from_index(efile, section->shdr->sh_link, &sec_str);
    if (sec_str.shdr == NULL) {
        warning("Couldn't get strings for GNU_verneed section at %#0" PRIx64 "\n", section->shdr->sh_offset);
        return;
    }

    print_header("Verdef: [%2u] '%s', n_ent: %d, Addr: %#010" PRIx64 ", Offset: %#08llx, Lnk: [%2u] '%s'\n",
	    (unsigned int) elf_ndxscn(section->scn),
	    elf_strptr(efile->elf, efile->shstrndx, section->shdr->sh_name),
            section->shdr->sh_info,
	    section->shdr->sh_addr,
	    section->shdr->sh_offset,
	    (unsigned int) section->shdr->sh_link,
	    elf_strptr(efile->elf, efile->shstrndx, sec_str.shdr->sh_name));


    UElf_Verdef_Iter iter = UElf_Verdef_Iter_INIT;
    while (iter_verdef(efile, section, &iter)) {
        unsigned int cnt = 0;
        while (iter_verdefaux(efile, section, &iter)) { 
            if (cnt == 0)
                print_verdef(efile, section, iter.offset, iter.def, iter.aux);
            else
                print_verdefaux(efile, section, iter.auxoffset, iter.aux);

            cnt++;
        };
    }
}

void show_versioninfo(UElf_File *efile) {
    UElf_Section section = UElf_Section_INIT;
    while (section_find(efile, SHT_GNU_versym, &section))
        handle_versym(efile, &section);

    section = UElf_Section_INIT;
    while (section_find(efile, SHT_GNU_verneed, &section))
        handle_verneed(efile, &section);

    section = UElf_Section_INIT;
    while (section_find(efile, SHT_GNU_verdef, &section))
        handle_verdef(efile, &section);
}
