#!/usr/bin/python2.4
# report.py - Generate a nice report from checklib logs
# Copyright (C) 2006  Christian Aichinger <Greek0@gmx.net>
# 
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 2 of the License.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
# 
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

import sys
import os
import re
import shutil

from kid import Template

import Image
import ImageDraw
import ImageFont

import pkg_tools.pkg_tools as pkg_tools
import config
from checklib import db_backend, flags

LOGDIR = config.Logdir
INDIR = config.Templatedir
OUTDIR = config.Reportdir

logcolors = {
    'OK': '#0f0',
    'PROBLEM': '#f00',
    'NOTICE': '#ff0',
    'SKIP': '#0c0',
    'NONE': '#fff',
    'ERROR': '#00e' }

def get_results():
    db_backend.init_database(config.Database)
    results = db_backend.get_results()

    for res in results:
        res.logfile = 'log.%s_%s.html' % (res.res_pkg_name, res.res_pkg_ver)

    return results

def draw_stats(size, stats, order):
    outfile = os.path.join(OUTDIR, 'graph.png')
    fontsize = 14
    text_align = 40

    im = Image.new("RGB", (size+150,size), '#ffffff')
    draw = ImageDraw.Draw(im)
    font = ImageFont.truetype('/usr/share/fonts/truetype/ttf-bitstream-vera/Vera.ttf', fontsize)
    
    offset = -90
    total = stats["TOTAL"]
    text_offset = 0
    for name in order:
        val = stats[name]
        name = name.split(',', 1)[0]
        if name in ['TOTAL', 'Internal problems']:
            continue

        pie_size = 360.0 * val / total
        draw.pieslice((0, 0, size-1, size-1), 
                int(round(offset)), 
                int(round(offset+pie_size)), 
                fill=logcolors[name], outline='#000000')
        offset += pie_size

        vsize = draw.textsize(name, font=font)[1]
        text_offset += vsize + 5

        draw.rectangle(
                (size+text_align+2, text_offset+2, size+text_align+vsize-2, text_offset+vsize-2),
                outline = '#000000', fill = logcolors[name])
        draw.text((size+text_align+vsize+5, text_offset), name, font=font, fill = '#000000')

    del draw 
    im.save(outfile, "PNG")

def start_of_name(pkgname):
    if pkgname.startswith('lib') and len(pkgname) >=4:
        return pkgname[:4]
    return pkgname[0]

def maint_from_result(pkgfile, res):
    try: maint = pkgfile[res.res_pkg_name]["Maintainer"]
    except KeyError:
        print "  Warning: pkg not in Packages any more, skipping:", res.res_pkg_name
        return None

    maint = ''.join([c for c in maint if not c in '"'])
    groups = re.search(r'^([^<]+?)\s*<([^>]+)>', maint).groups()
    maint = ' - '.join(reversed(groups))

    try: maint = unicode(maint, "utf-8")
    except UnicodeDecodeError: maint = unicode(maint, "latin1")

    return maint

def main():
    print "Getting results from the database"
    results = get_results()


    print "Parsing 'Packages' file"
    try:
        pkgfile = pkg_tools.PackagesFile(config.pkg_file_name)
    except IOError:
        tmppath = os.path.join(config.Cache, 'Packages')
        pkgfile = pkg_tools.PackagesFile(tmppath)


    print "Generating report (1)"

    res_order = [
        (flags.RES_OK, 'OK'),
        (flags.RES_SKIP, 'SKIP'),
        (flags.RES_WARN, 'NOTICE'),
        (flags.RES_ERR, 'PROBLEM'),
        (flags.RES_ABORT, 'ERROR'),
    ]
    res_names = dict(res_order)
    res_order = [name for res,name in res_order]
    display_order = ['TOTAL'] + res_order

    results_by_res = {}
    results_by_name = {}
    results_by_maint = {}
    for res in results:
        maint = maint_from_result(pkgfile, res)
        if not maint:
            continue

        res.res_result = flags.severity(res.res_result)
        res.res_result_str = res_names[res.res_result]
        res.maint = maint

        results_by_res.setdefault(res.res_result, []).append(res)
        results_by_name.setdefault(start_of_name(res.res_pkg_name).upper(), []).append(res)
        results_by_maint.setdefault(maint[0].upper(), []).append(res)


    stats = {}
    stats['TOTAL'] = len(results)
    for severity, name in res_names.items():
        stats[name] = len(results_by_res[severity])

    print "Writing text file"

    f = file(os.path.join(OUTDIR, 'result.txt'), 'w')
    for res in results:
        if hasattr(res, 'res_result_str'):
            f.write('%s_%s %s %s %s\n' % (res.res_pkg_name, res.res_pkg_ver, res.res_result_str, res.res_flags,
                    'http://rerun.lefant.net/checklib/log.%s_%s.txt' % (res.res_pkg_name, res.res_pkg_ver)))
    f.close()


    print "Writing reports by result"

    for severity in [flags.RES_WARN, flags.RES_ERR, flags.RES_ABORT]:
        name = res_names[severity]
        fname = 'status.%s.html' % name
        fname2 = 'status.%s.frames.html' % name
        kidfile = os.path.join(INDIR, 'packages.kid')
        outfile = os.path.join(OUTDIR, fname)
        outfile2 = os.path.join(OUTDIR, fname2)

        template = Template(file=kidfile, results=results_by_res[severity],
                            logcolors=logcolors, char=name, INDIR=INDIR,
                            sortedby="status", fname=fname)
        template.write(outfile, output='xhtml-strict')

        assert 0 == os.system("""sed -e 's/"_self"/"chkl_bottom"/g' < '%s' > '%s'""" % (outfile,outfile2))



    print "Writing reports by package name"

    for char in sorted(results_by_name.keys()):
        print "  Writing char '%s'" % char

        fname = 'packages.%s.html' % char
        fname2 = 'packages.%s.frames.html' % char
        kidfile = os.path.join(INDIR, 'packages.kid')
        outfile = os.path.join(OUTDIR, fname)
        outfile2 = os.path.join(OUTDIR, fname2)

        template = Template(file=kidfile, results=results_by_name[char],
                            logcolors=logcolors, char=char, INDIR=INDIR,
                            sortedby="package name", fname=fname)
        template.write(outfile, output='xhtml-strict')

        assert 0 == os.system("""sed -e 's/"_self"/"chkl_bottom"/g' < '%s' > '%s'""" % (outfile,outfile2))



    print "Writing reports by maintainer email"

    for char in sorted(results_by_maint.keys()):
        print "  Writing char '%s'" % char

        maintainers = {}
        for res in results_by_maint[char]:
            maintainers.setdefault(res.maint, []).append(res)

        fname = 'maintainers.%s.html' % char
        fname2 = 'maintainers.%s.frames.html' % char
        kidfile = os.path.join(INDIR, 'maintainers.kid')
        outfile = os.path.join(OUTDIR, fname)
        outfile2 = os.path.join(OUTDIR, fname2)

        template = Template(file=kidfile, char=char,
                maintainers=maintainers, logcolors=logcolors, INDIR=INDIR,
                fname=fname)
        template.write(outfile, encoding="utf-8", output='xhtml-strict')

        assert 0 == os.system("""sed -e 's/"_self"/"chkl_bottom"/g' < '%s' > '%s'""" % (outfile,outfile2))
        

    print "Writing report index"

    # Make nice graphic
    draw_stats(180, stats, display_order)

    # Generate index
    kidfile = os.path.join(INDIR, 'index.kid')
    outfile = os.path.join(OUTDIR, 'index.html')
    outfile2 = os.path.join(OUTDIR, 'index.frames.html')

    pkg_chars = sorted([(len(char), char) for char in results_by_name.keys()])
    pkg_chars = [char for l, char in pkg_chars]
    maint_chars = sorted(results_by_maint.keys())

    template = Template(file=kidfile, stats=stats, logcolors=logcolors,
                        maint_chars = maint_chars, pkg_chars = pkg_chars, INDIR=INDIR,
                        stats_order = display_order, frmurl = '')
    template.write(outfile, output='xhtml-strict')

    template = Template(file=kidfile, stats=stats, logcolors=logcolors,
                        maint_chars = maint_chars, pkg_chars = pkg_chars, INDIR=INDIR,
                        stats_order = display_order, frmurl = '.frames')
    template.write(outfile2, output='xhtml-strict')

    print "Generating rst pages"

    assert 0 == os.system("rst2html --initial-header-level=2 '%s' '%s'" % (
        os.path.join(INDIR, 'howto-fix-problems.rst'),
        os.path.join(OUTDIR, 'howto-fix-problems.html')))
    assert 0 == os.system("rst2html --initial-header-level=2 '%s' '%s'" % (
        os.path.join(INDIR, 'howto-fix-errors.rst'),
        os.path.join(OUTDIR, 'howto-fix-errors.html')))


    print "Copying files"

    shutil.copy(os.path.join(INDIR, 'frames.html'),
                os.path.join(OUTDIR, 'frames.html'))


if __name__=='__main__':
    main()
